#include <math.h>

#define NAME_LENGTH 40
#define LINE_LENGTH 100
#define LONG_LINE_LENGTH 	512


/* OPTIONS */

#define M_SET    0x0001
#define JM_SET   0x0002
#define H_SET    0x0004
#define JH_SET   0x0008
#define S_SET    0x0010
#define JS_SET   0x0020
#define R_SET    0x4000
#define JR_SET   0x8000
#define J_SET   (JM_SET | JH_SET | JS_SET | JR_SET)

#define Q_SET    0x0040
#define V_SET    0x0080

#define P_SET    0x0100

#define C_SET    0x0400
#define B_SET    0x0800
#define TWO_SET  0x1000

#define ID_SET	0x00010000
#define IH_SET  0x00020000
#define IS_SET  0x00040000

#define W_SET   0x00080000

#define BATCH_SET  0x00080000
#define I_SET   (ID_SET | IH_SET | IS_SET)

#define LATCH_SET 0x00100000


#define COLLECT_PHASE   0
#define COLLECT_VOLTAGE 1

#define COLLECT_BY_PHASE   0
#define COLLECT_BY_VOLTAGE 1
#define COLLECT_EDGES      2

#define STEP_FORWARD 5

#define RISING_EDGE  0
#define FALLING_EDGE 1


typedef float TIME;
typedef float PHASE;
typedef float VOLTAGE;

#define MIN(x,y) (((x)<(y))?(x):(y))
#define MAX(x,y) (((x)<(y))?(y):(x))


typedef struct
{
 /* maximum number of hull planes */
 int maxplane;

 /* maximum number of iterations */
 int maxiter;

 /* maximum number of iterations */
 int miniter;

 /* pulse extraction method
     possible values:
       COLLECT_BY_PHASE    = phase threshold method
       COLLECT_BY_VOLTAGE  = peak voltage method 
 */
 int pulse_extraction_method;

 /* phase threshold that determines the exact moment
    when the pulse apears */
 PHASE nom_phase_thresh;

 /* minimum expected value of the phase in simulated timepoints
    closest to the phase threshold */
 PHASE min_phase_thresh;

 /* maximum expected value of the phase in simulated timepoints
    closest to the phase threshold */
 PHASE max_phase_thresh;

 /* if this variable is set to 1 the time when the value of
    the signal phase crosses the threshold is determined
    by the use of interpolation;
    if this variable is set to 0 the simulation point
    with the value of phase closest to the threshold
    is used;
 */
 int interpolate;

 /* the minimum value of voltage peak that can be treated 
    as the RSFQ pulse */
 VOLTAGE volt_thresh;

 /* the minimum value of difference in phase between 
    two consucutive voltage peaks above which they are treated
    as two separate pulses;
    if the difference in phase is less than minimum
    the peak with higher voltage level is assumed to be the pulse,
    the other is ignored */
 PHASE min_pulse_phase_diff;

 /* ratio determining the timepoint within the clock cycle for which
    the passfail test for proper phase is performed; */
 float checkpoint_ratio;

 /* minimum difference between real and expected value of phase
    threshold considered as test failure */
 PHASE min_fail_thresh;

/* accuracy in binary search for hold and setup times */
  TIME accuracy;

/* timing search accuracy */
  TIME time_accuracy;

/* maximum assumed variation of the output propagation delay resulting
   from variation of circuit operating parameters */
  TIME max_delay_var;

/* maximum assumed variation of the index corresponding to 
   max_delay_var in delay */
  int max_index_var;

/* min nr of simulation steps between passfail checktimes */
  int min_time_step;

/* marker, if inputs are to be checked for extra pulses */
  int inputs_checked;

/* names of subdirectories: */
 /* for input files */
 char input_dir[NAME_LENGTH];

 /* for temporary files */
 char tmp_dir[NAME_LENGTH];

 /* for output files */
 char output_dir[NAME_LENGTH];

/* call name for jspice: */
 char spice_name[NAME_LENGTH];

/* extensions: */
 char circuit_ext[NAME_LENGTH];
 char param_ext[NAME_LENGTH];
 char print_ext[NAME_LENGTH];
 char passfail_ext[NAME_LENGTH];
 char phase_ext[NAME_LENGTH];
 char step_ext[NAME_LENGTH];
 char check_ext[NAME_LENGTH];
 char avr_ext[NAME_LENGTH];
 char hold_ext[NAME_LENGTH];
 char setup_ext[NAME_LENGTH]; 
 char separation_ext[NAME_LENGTH];
 char good_ext[NAME_LENGTH]; 
 char bad_ext[NAME_LENGTH];
 char logic_ext[NAME_LENGTH]; 
 char async_ext[NAME_LENGTH];
 char hose_ext[NAME_LENGTH];
 char report_ext[NAME_LENGTH];
 char period_ext[NAME_LENGTH];
 char comp_ext[NAME_LENGTH];
 char nom_ext[NAME_LENGTH];
 char init_ext[NAME_LENGTH];
 char marg_ext[NAME_LENGTH];
 char opt_ext[NAME_LENGTH];
 char bound_ext[NAME_LENGTH];
 char region_ext[NAME_LENGTH];
 char timeinit_ext[NAME_LENGTH];
 char signal_ext[NAME_LENGTH];

 /* circuit name */
 char circuit_name[NAME_LENGTH];

 /* code of the program which is run;
    used to modify the function of
    the procedures called from the different 
    programs of the packet;
    possible values of this variable are
    given below;
 */
 int program;

 /* parameter set number */
 char opt_num[NAME_LENGTH];

 /* command line options */
 int input_options;

/* OPTIONS FOR LATCHING LOGIC */

/* average voltage in the "1" state */
  double avr_one_volt;

/* thresholds used to determine the position of the rising and falling
   edge of the signal in MVTL logic */
  double nom_edge_thresh;

/* maximum voltage in the "0" state */
  double max_zero_volt;

/* minimum voltage in the "1" state */
  double min_one_volt;

/* minimum separation between rising and falling edge of the signal
   in the latching logic */
  TIME min_edge_sep;

/* minimum interval between passfail checks */
  TIME min_checks_interval;

/* minimum interval between checkpoint and the rising edge of the signal
   for nominal values of parameters */
  TIME min_interval_to_riseedge;

/* minimum interval between the falling edge of the signal
   and the checkpoint for nominal values of parameters */
  TIME min_interval_to_falledge;

/* ratio determining the timepoint within an interval 
   of high value of the clock signal, when the output signals
   are checked for the correct value */
  double clk_chktime_ratio;

} PARAMETERS;

extern PARAMETERS param;


/* possible values of the field program in the structure PARAMETERS */

#define	HS_PROGRAM 	0
#define INIT_PROGRAM	1
#define MARG_PROGRAM	2
#define OPT_PROGRAM	3

typedef struct
{
 /* minimum value of the generator parameter describing position of data
    pulses within clock pulses */
 TIME xmin;

 /* maximum value of the generator parameter describing position of data
    pulses within clock pulses */
 TIME xmax;

 /* maximum value of the generator parameter describing position of data
    pulses within clock pulses */
 TIME x;

 /* time step between subsequent simulations */
 TIME step;

 /* number of input driven by the variable generator */
 int input_nr;

 /* number of the input clock cycle for which the propagation delay
    is computed */
 int in_clk_cycle;

 /* number of the output clock cycle for which the propagation delay
    is computed */
 int out_clk_cycle;


} XY_PARAMETERS;


struct PULSE_DESC
{
/* values describing the pulse */

 TIME t;          /* position in time */
 int index;       /* # of sample */

 double val;      /* voltage */
 PHASE phase;     /* phase */

/* the following fields are used only when the pulse is extracted by phase: */
 PHASE norm_phase;   /* phase rounded to k*2*PI */
 TIME t_below;
 TIME t_above;
 PHASE p_below;
 PHASE p_above;

/* pointers to the next and previous positions on the list */
 struct PULSE_DESC* next;
 struct PULSE_DESC* prev;
};

typedef struct PULSE_DESC PULSE;


struct EDGE_DESC
{
/* values describing the pulse */

 TIME t;          /* position in time */
 int index;       /* # of sample */

 double val;      /* voltage */
 int direction;   /* edge type: 0 = rising; 1 = falling; */

/* the more exact position of the pulse in time obtained by interpolation */
 TIME exact_time; 

/* pointers to the next and previous positions on the list */
 struct EDGE_DESC* next;
 struct EDGE_DESC* prev;
};

typedef struct EDGE_DESC EDGE;


typedef struct
{
/* logical value of the signal in a given clk cycle */
 int val;

/* distance between the input/output pulse and the preceeding clk pulse */
 TIME *delta;

} LOGIC;



struct async_checkpoint
{
/* output nr */
 int sig_nr;

/* point type = initial point or check point */
 int type;

/* index corresponding to the checkpoint */
 int index;

/* proper phase difference at the given output divided by pi */
 int pi_phase;

/* pointer to the next element of the list */
 struct async_checkpoint * next;
};
typedef struct async_checkpoint ASYNC_CHECK;


/* possible values of the field type */
#define REGULAR_CHECKPOINT  0
#define INIT_CHECKPOINT     1


typedef struct
{
/* number of signals traced */
 int     signal_num;

/* number of clk signals traced */
 int     clk_num;

/* number of input signals traced */
 int     ins_num;

/* number of internal signals traced */
 int     inter_num;

/* number of output signals traced */
 int     outs_num;

/* names of the signals given by the designer */
 char  (*signal_names)[NAME_LENGTH];

/* names of the signals used in jspice simulations */
 char  (*jspice_voltage_names)[NAME_LENGTH];

/* names of the signals used in jspice simulations */
 char  (*jspice_phase_names)[NAME_LENGTH];

/* number of the input with variable generator attached */
 int   var_in_nr;

/* time resolution (distance in time between two adjecant input samples) */
 TIME  min_time_delta;

/* time at which simulation tracing begins */
 TIME start_time;

/* number of input samples available */
 int   indices_nr;


/* array (of signal_num length) of pointers to the begining
   of list of pulses (each pulse represented by the PULSE structure) */
 PULSE **sig_bgn;

/* array (of signal_num length) of pointers to the current end
   of list of pulses (each pulse represented by the PULSE structure) */
 PULSE **sig_end;

/* array (of signal_num length) of pointers to the begining
   of list of edges (each edge represented by the EDGE structure) */
 EDGE **sig_edge_bgn;

/* array (of signal_num length) of pointers to the current end
   of list of edges (each edge represented by the EDGE structure) */
 EDGE **sig_edge_end;

/* array (of signal_num length) of initial logic values
   of the signals in latching logic*/
 int *sig_initial;

/* array (of signal_num length) of the current phases
   for each signal (rounded to k*2*PI) */
 PHASE   *current_phase;

/* array (of signal_num length) of indices of the last samples 
   greater than the voltage threshold */
 int   *last_index;

/* array of points in time corresponding to clk pulses */
 TIME   *clk_times;

/* number of clk cycles simulated */
 int     clk_cycles_nr;

/* array (of signal_num length) of pointers to the beginnings
   of arrays (of clk_cycles_nr length) of logic values in each traced node 
   (logic value in each clk cycle represented by the LOGIC structure) */
 LOGIC **sig_log;

/* array (of clk_cycles_nr length) of indices for which pass/fail 
   check is performed in passfail JSPICE control file; 
   they are chosen as laying in a given point of each clk_cycle; */
 int    *check_index;

/* array (of signal_num length) of indices corresponding to the points
   where phase is sampled at the beginning of the simulation interval 
   being traced */
 int *initial_phase_index;

/* array (of outs_num in length) of pointers to lists 
   of indices for which pass/fail check is performed in passfail 
   JSPICE control file; 
   the indices correspond to timepoints MAX_DELAY_VARIATION before and after 
   each output pulse; */
 ASYNC_CHECK   **async_check_index;

 ASYNC_CHECK   *async_check_list;

} SIGNALS;


typedef struct
{ 
/* hold or setup time */
 TIME delta;

/* clk cycle in which hold/setup requirement is violated */
 int cycle;

/* # of input at which hold/setup requirement is violated */
 int input;

} HOLDSETUP;



typedef union
{
 struct
 {
  TIME min;
  TIME avr;
  TIME max;
 } t;
 struct
 {
  PHASE min;
  PHASE avr;
  PHASE max;
 } p;
 struct
 {
  VOLTAGE min;
  VOLTAGE avr;
  VOLTAGE max;
 } v;
}
MIN_MAX_AVR;


struct DELAYS_DESC
{
 TIME delta;
 int clk_cycle;
 struct DELAYS_DESC *next;
};

typedef struct DELAYS_DESC DELAYS;

struct INPUTS_DESC
{
 int *input_val;
 DELAYS *delay;
 DELAYS *last_delay;
 struct INPUTS_DESC *next;
};

typedef struct INPUTS_DESC INPUTS;


typedef struct
{
 double val;
 int sgn;
 int param_nr;
 int pseudo;
 double error;
} MARGIN;


struct PARAMETER_SET_STRUCT
{
 int x;
 int y;
 int points_nr;
 struct PARAMETER_SET_STRUCT *next;
};

typedef struct PARAMETER_SET_STRUCT PARAMETER_SET;


struct TIME_PARAM_STRUCT
{
 int param_nr;
 char choice_meth;
 double start_mean_value;
 double end_deviation_value;
 double delta_ratio;
 double value;
 int points_nr;
 struct TIME_PARAM_STRUCT *next;
};

typedef struct TIME_PARAM_STRUCT TIME_PARAM;

/* possible values of the field choice_meth in the TIME_PARAM_STRUCT */
#define ADD_DELTA	'A'
#define MUL_BY_FACTOR	'M'
#define RANDOM_UNIFORM  'R'
#define RANDOM_NORMAL	'N'


struct SYNC_TIME_DELAY_STRUCT
{
 int sig_num;
 int clk_cycle;
 struct SYNC_TIME_DELAY_STRUCT *next;
};

typedef struct SYNC_TIME_DELAY_STRUCT SYNC_TIME_DELAY;


struct ASYNC_TIME_DELAY_STRUCT
{
 int sig1_num;
 int pulse1_num;
 int sig2_num;
 int pulse2_num;
 struct ASYNC_TIME_DELAY_STRUCT *next;
};

typedef struct ASYNC_TIME_DELAY_STRUCT ASYNC_TIME_DELAY;



